from typing import Optional, List
from dataclasses import dataclass

from .creative_center_data_types import Season, Sections, Section, SeasonEpisodeFullDetail, VideoArchive, VideoPage


@dataclass
class GetVideoResponse:
    archive: VideoArchive
    videos: List[VideoPage]

@dataclass
class GetSeasonResponse:
    season: Season
    course: Optional[None]
    checkin: Optional[None]
    seasonStat: Optional[None]
    sections: Sections
    part_episodes: Optional[List[SeasonEpisodeFullDetail]]

@dataclass
class GetSectionResponse:
    section: Section
    episodes: Optional[List[SeasonEpisodeFullDetail]]


from typing import List, Tuple
from dataclasses import dataclass

from .base import BaseClient, PaginationData
from .space_data_types import SeriesVideoInfo


class BilibiliSpace(BaseClient):
    def get_relevant_videos_for_series(self, mid: int, series_id: int, keywords: str = "") -> List[SeriesVideoInfo]:
        resp = self._get_request(url="https://api.bilibili.com/x/series/recArchivesByKeywords", params={"mid": mid, "keywords": keywords, "series_id": series_id})
        if resp.status_code >= 400:
            return []
        resp_json = resp.json()
        if resp_json["code"] != 0:
            return []
        return [SeriesVideoInfo(**video) for video in resp_json["data"]["archives"]]

    def add_video_to_series(self, series_id: int, aid_list: List[int]) -> Tuple[bool, str]:
        resp = self._post_request(url="https://api.bilibili.com/x/series/series/addArchives",
                                  params={"csrf": self.credential.csrf},
                                  data={
                                      "mid": self.credential.uid,
                                      "series_id": series_id,
                                      "aids": ",".join(map(lambda x: str(x), aid_list)),
                                  })
        if resp.status_code >= 400:
            self.logger.error(f"Failed to add video to series, status code {resp.status_code}")
            return False, f"Failed to add video to series, status code {resp.status_code}"
        resp_json = resp.json()
        if resp_json["code"] != 0:
            self.logger.error(f"Failed to add video to series, response {resp_json['message']}")
            return False, resp_json["message"]
        return True, ""


@dataclass
class GetRelevantVideoForSeriesResponse:
    archives: List[SeriesVideoInfo]
    page: PaginationData


from typing import Union
from dataclasses import dataclass
import logging
import os

import requests


REQUEST_HEADERS = {
    "Referer": "https://www.bilibili.com/",
    "Connection": "keep-alive",
    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36"
}
REQUEST_TIMEOUT = 30

class BilibiliCredential:
    def __init__(self,
                 uid: Union[str, int],
                 sessdata: str, csrf: str,
                 username: str = None, password: str = None,
                 buvid3: str = None, access_token: str = None):
        self.uid: str = str(uid)
        self.username = username
        self.password = password
        self.sessdata = sessdata
        self.csrf = csrf
        self.buvid3 = buvid3
        self.access_token = access_token

    @staticmethod
    def init_from_env(raise_if_invalid: bool = True):
        print("Initializing credential from environment variables: BILIBILI_UID, BILIBILI_SESSDATA, BILIBILI_CSRF")
        instance = BilibiliCredential(uid=os.environ.get("BILIBILI_UID"), sessdata=os.environ.get("BILIBILI_SESSDATA"), csrf=os.environ.get("BILIBILI_CSRF"))
        if raise_if_invalid and not instance.validate():
            raise ValueError("Invalid credential")
        return instance

    @staticmethod
    def is_not_empty(input_value: str) -> bool:
        return input_value is not None and input_value != ""

    def get_cookie(self) -> dict:
        return_dict = {}
        if self.is_not_empty(self.sessdata):
            return_dict["SESSDATA"] = self.sessdata
        if self.is_not_empty(self.csrf):
            return_dict["bili_jct"] = self.csrf
        if self.is_not_empty(self.buvid3):
            return_dict["buvid3"] = self.buvid3
        return return_dict

    def get_cookie_for_url(self, url: str) -> dict:
        if url.find(".bilibili.com") >= 0:
            return self.get_cookie()
        else:
            return {}

    def validate(self) -> bool:
        print("Validating credential...")
        resp = requests.get("https://space.bilibili.com/", headers=REQUEST_HEADERS, cookies=self.get_cookie())
        print(resp.status_code, resp.url)
        if (resp.url == "https://passport.bilibili.com/login?gourl=https://space.bilibili.com"
                or resp.url == "https://passport.bilibili.com/pc/passport/login?gourl=https%3A%2F%2Fspace.bilibili.com"):  # will try to redirect to login page if SESSDATA is expired
            print("bilibili.com credential not valid", self.get_cookie())
            return False
        print("bilibili.com credential valid")
        return True

class BaseClient:
    def __init__(self, credential: BilibiliCredential,
                 logger: logging.Logger = logging.getLogger("bilibili_api_c")
                 ):
        self.credential = credential
        self.requests_session = requests.Session()
        self.logger = logger

    def _get_request(self, url: str, params: dict, use_credential: bool = True) -> requests.Response:
        return self.requests_session.get(url=url, params=params, cookies=self.credential.get_cookie_for_url(url) if use_credential else {}, headers=REQUEST_HEADERS)

    def _post_request(self, url: str, params: dict, data: dict, use_credential: bool = True) -> requests.Response:
        return self.requests_session.post(url=url, params=params, data=data, cookies=self.credential.get_cookie_for_url(url) if use_credential else {}, headers=REQUEST_HEADERS)


@dataclass
class PaginationData:
    num: int
    size: int
    total: int


from typing import Optional, Generator, List, Tuple
from enum import Enum

from .base import BaseClient
from .creative_center_data_types import BilibiliVideoDetailCreativeCenter, SectionEpisode, EpisodeSorts, SectionSorts
from .creative_center_response_types import GetSeasonResponse, GetSectionResponse, GetVideoResponse


class EpisodesSortMethod(Enum):
    PUBLISH_ASC = "asc"
    PUBLISH_DESC = "desc"
    AS_IS = "as_is"

class BilibiliCreativeCenter(BaseClient):
    def get_video_detail(self, aid: int) -> Optional[GetVideoResponse]:
        resp = self._get_request(url="https://member.bilibili.com/x/web/archive/videos", params={"aid": aid})
        if resp.status_code >= 400:
            self.logger.error(f"got status code {resp.status_code} from {resp.url}")
            return None

        resp_json = resp.json()
        if resp_json["code"] != 0:
            self.logger.error(f"got error code {resp_json['code']} {resp_json['message']}")
            return None

        return GetVideoResponse(**resp_json["data"])

    def get_videos_by_page(self, page_number: int = 1) -> Optional[list[BilibiliVideoDetailCreativeCenter]]:
        resp = self._get_request(url="https://member.bilibili.com/x/web/archives",
                                 params={"status": "pubed", "pn": page_number, "ps": 10, "coop": 1, })
        if resp.status_code >= 400:
            self.logger.error(f"got status code {resp.status_code} from {resp.url}")
            return None

        try:
            response_json = resp.json()
        except ValueError:
            self.logger.error("unable to decode json")
            return None

        print(response_json)
        if response_json["code"] != 0:
            self.logger.error(f"got error code {response_json['code']} {response_json['message']}")
            return None

        if response_json["data"]["arc_audits"] is None:
            self.logger.error("arc_audits is null")
            return None

        return_list: list[BilibiliVideoDetailCreativeCenter] = []
        try:
            for each_video in response_json["data"]["arc_audits"]:
                # print(each_video)
                return_list.append(BilibiliVideoDetailCreativeCenter(
                    aid=each_video["Archive"]["aid"],
                    bvid=each_video["Archive"]["bvid"],
                    title=each_video["Archive"]["title"],
                    description=each_video["Archive"]["desc"],
                    tags=each_video["Archive"]["tag"],
                    duration=each_video["Archive"]["duration"],
                    copyright=each_video["Archive"]["copyright"],
                    source=each_video["Archive"]["source"],
                    ptime=each_video["Archive"]["ptime"],
                    ctime=each_video["Archive"]["ctime"],
                    zone_id=each_video["Archive"]["tid"],
                    zone_name=each_video["typename"],
                    subtitle_count=each_video["captions_count"],
                ))
            # END for
        except KeyError:
            self.logger.error("unable to find parse arc_audits (KeyError)")
            return None
        return return_list

    def get_videos_generator(self, page_limit: int = 10) -> Generator[BilibiliVideoDetailCreativeCenter, None, None]:
        """
        Get videos from the creative center

        :param page_limit: -1 for no limit
        :return:
        """
        current_page_number = 1
        while current_page_number <= page_limit or page_limit == -1:
            video_list = self.get_videos_by_page(page_number=current_page_number)
            if video_list is None:
                self.logger.error(f"Failed to get video list from page {current_page_number}")
                return
            for video in video_list:
                yield video
            # END for
            self.logger.info(f"Page {current_page_number} done")
            current_page_number += 1
        # END while

    def get_season_detail(self, season_id: int) -> Optional[GetSeasonResponse]:
        resp = self._get_request(url="https://member.bilibili.com/x2/creative/web/season", params={"id": season_id})
        if resp.status_code >= 400:
            self.logger.error(f"got status code {resp.status_code} from {resp.url}")
            return None
        resp_json = resp.json()
        if resp_json["code"] != 0:
            self.logger.error(f"got error code {resp_json['code']} {resp_json['message']}")
            return None

        return GetSeasonResponse(**resp.json())

    def get_section_detail(self, section_id: int) -> Optional[GetSectionResponse]:
        resp = self._get_request(url="https://member.bilibili.com/x2/creative/web/section", params={"id": section_id})
        if resp.status_code >= 400:
            self.logger.error(f"got status code {resp.status_code} from {resp.url}")
            return None
        resp_json = resp.json()
        if resp_json["code"] != 0:
            self.logger.error(f"got error code {resp_json['code']} {resp_json['message']}")
            return None

        return GetSectionResponse(**resp.json())

    def add_episodes_to_section(self, section_id: int, episodes: List[SectionEpisode]):
        resp = self._post_request(url="https://member.bilibili.com/x2/creative/web/season/section/episodes/add",
                                  params={"csrf": self.credential.csrf},
                                  data={
                                      "sectionId": section_id,
                                      "episodes": episodes,
                                  })
        if resp.status_code >= 400:
            self.logger.error(f"got status code {resp.status_code} from {resp.url}")
            return False, f"got status code {resp.status_code} from {resp.url}"
        resp_json = resp.json()
        if resp_json["code"] != 0:
            self.logger.error(f"got error code {resp_json['code']} {resp_json['message']}")
            return False, resp_json["message"]

    def submit_section_changes(self, section_detail: GetSectionResponse, sort: EpisodesSortMethod = EpisodesSortMethod.AS_IS) -> Tuple[bool, str]:
        episode_sorts: List[EpisodeSorts] = []
        match sort:
            case EpisodesSortMethod.AS_IS:
                for episode in section_detail.episodes:
                    episode_sorts.append(EpisodeSorts(id=episode.id, sort=episode.order))
            case EpisodesSortMethod.PUBLISH_ASC:
                section_detail.episodes.sort(key=lambda x: x.aid)
                for i, episode in enumerate(section_detail.episodes):
                    episode_sorts.append(EpisodeSorts(id=episode.id, sort=i+1))
            case EpisodesSortMethod.PUBLISH_DESC:
                section_detail.episodes.sort(key=lambda x: x.aid, reverse=True)
                for i, episode in enumerate(section_detail.episodes):
                    episode_sorts.append(EpisodeSorts(id=episode.id, sort=i+1))
            # END match
        # END match

        resp = self._post_request(url="https://member.bilibili.com/x2/creative/web/season/section/edit",
                                  params={"csrf": self.credential.csrf},
                                  data={
                                      "section": {
                                          "id": section_detail.section.id,
                                          "seasonId": section_detail.section.seasonId,
                                          "title": section_detail.section.title,
                                          "type": section_detail.section.type,
                                      }
                                  })
        if resp.status_code >= 400:
            self.logger.error(f"got status code {resp.status_code} from {resp.url}")
            return False, f"got status code {resp.status_code} from {resp.url}"
        resp_json = resp.json()
        if resp_json["code"] != 0:
            self.logger.error(f"got error code {resp_json['code']} {resp_json['message']}")
            return False, resp_json["message"]

        return True, "success"

    def submit_season_changes(self, season_detail: GetSeasonResponse) -> Tuple[bool, str]:
        section_sorts: List[SectionSorts] = []
        for section in season_detail.sections.sections:
            section_sorts.append(SectionSorts(id=section.id, sort=section.order))

        resp = self._post_request(url="https://member.bilibili.com/x2/creative/web/season/edit",
                                  params={"csrf": self.credential.csrf},
                                  data={
                                      "season": {
                                          "cover": season_detail.season.cover,
                                          "desc": season_detail.season.desc,
                                          "id": season_detail.season.id,
                                          "isEnd": season_detail.season.isEnd,
                                          "season_price": season_detail.season.season_price,
                                          "title": season_detail.season.title,
                                      },
                                      "sorts": section_sorts,
                                  })
        if resp.status_code >= 400:
            self.logger.error(f"got status code {resp.status_code} from {resp.url}")
            return False, f"got status code {resp.status_code} from {resp.url}"
        resp_json = resp.json()
        if resp_json["code"] != 0:
            self.logger.error(f"got error code {resp_json['code']} {resp_json['message']}")
            return False, resp_json["message"]

        return True, "success"




from typing import Optional, List
from dataclasses import dataclass

@dataclass
class VideoArchive:
    aid: int
    bvid: str
    title: str

@dataclass
class VideoPage:
    cid: int
    index: int
    title: str
    duration: int


@dataclass
class BilibiliVideoDetailCreativeCenter:
    bvid: str
    aid: int
    title: str
    description: str
    tags: list[str]
    duration: int
    copyright: int
    source: str
    zone_id: int  # tid
    zone_name: str  # typename
    subtitle_count: int
    # date related
    ptime: int
    ctime: int

@dataclass
class Season:
    id: int
    title: str
    desc: str
    cover: str
    isEnd: int
    mid: int
    isAct: int
    is_pay: int
    state: int
    partState: int
    signState: int
    rejectReason: str
    ctime: int
    mtime: int
    no_section: int
    forbid: int
    protocol_id: str
    ep_num: int
    season_price: int
    is_opened: int
    has_charging_pay: int
    has_pugv_pay: int
    SeasonUpfrom: int

@dataclass
class SectionSorts:
    id: int  # episode id
    sort: int

@dataclass
class Section:
    id: int
    type: int
    seasonId: int
    title: str
    order: int
    state: int
    partState: int
    rejectReason: str
    ctime: int
    mtime: int
    epCount: int
    cover: str
    has_charging_pay: int
    Episodes: Optional[None]
    show: int
    has_pugv_pay: int

@dataclass
class Sections:
    sections: List[Section]
    total: int

@dataclass
class EpisodeSorts:
    id: int  # episode id
    sort: int

@dataclass
class SectionEpisode:
    aid: int
    cid: int
    title: str
    charging_pay: int = 0
    member_first: int = 0


@dataclass
class SeasonEpisodeFullDetail:
    id: int
    title: str
    aid: int
    bvid: str
    cid: int
    seasonId: int
    sectionId: int
    order: int
    videoTitle: str
    archiveTitle: str
    archiveState: int
    rejectReason: str
    state: int
    cover: str
    is_free: int
    aid_owner: bool
    charging_pay: int
    member_first: int
    pugv_pay: int



from typing import Optional
from dataclasses import dataclass

@dataclass
class SeriesVideoInfo:
    aid: int
    title: str
    pubdate: int
    ctime: int
    state: int
    pic: str
    duration: int
    stat: Optional[None]
    bvid: str
    ugc_pay: int
    interactive_video: bool
    enable_vt: int
    vt_display: str
    playback_position: int